﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Net;
using System.Threading;

namespace vcsharp_ethernet_client_server_demo
{
    public partial class Form1 : Form
    {
        private const Int32 bufferSize = 65535;
        private const Int32 timeoutMsec = 1500;
        private Int32 serverPortSelect = 0;
        private Int32 clientPortSelect = 0;
        private IPAddress ipAddressServerBind = null;
        private IPAddress ipAddressClientConnect = null;
        private TcpListener tcpListener = null;
        private TcpClient tcpClientOnServer = null;
        private Socket tcpClientSocket = null;
        private IAsyncResult clientSocketConnect = null;
        private IPEndPoint clientSocketEndPoint = null;
        private NetworkStream networkStreamConnectedClient = null;
        private Thread threadClientSide = null;
        private byte[] serverTransmitBuffer = new byte[bufferSize];
        private byte[] serverReceiveBuffer = new byte[bufferSize];
        private byte[] clientReceiveBuffer = new byte[bufferSize];
        private byte[] clientTransmitBuffer = new byte[bufferSize];
        private bool clientDisconnectCommand = false;
        private bool serverConnected = false;

        public Form1()
        {
            InitializeComponent();
        }

        bool ValidateIPAddress(string ipAddress)
        {
            char fullStop = '.';
            string[] octetBytes = null;
            int octetValue = 0;
            int loopCount = 0;
            octetBytes = ipAddress.Split(fullStop);
            /// Checks number of octets in IP address.
            if (octetBytes.Length != 4)
            {
                return false;
            }
            try
            {
                /// Checks maximum octet value i.e. 255.
                for (loopCount = 0; loopCount < 4; loopCount++)
                {
                    octetValue = Convert.ToInt32(octetBytes[loopCount]);
                    if (octetValue < 0 || octetValue > 255)
                    {
                        throw new ArgumentOutOfRangeException();
                    }
                }
            }
            catch
            {
                return false;
            }
            return true;
        }

        private void serverConnectButton_Click(object sender, EventArgs e)
        {
            string hostName = null;
            bool success = false;
            int addressCount = 0;
            if (serverConnectButton.Text == "Listen")
            {
                try
                {
                    /// Checks if port number is valid.
                    serverPortSelect = (Int32)Convert.ToUInt32(serverPortNumber.Text);
                    if ((serverPortSelect > 65535) || (serverPortSelect < 1))
                    {
                        throw new ArgumentOutOfRangeException();
                    }
                }
                catch
                {
                    serverConnectionDetails.Text = "Port number must be positive value less than 65535.";
                    return;
                }
                serverPortNumber.Enabled = false;
                serverConnectButton.Text = "Trying...";
                serverConnectButton.Enabled = false;
                serverConnectionDetails.Text = "Trying to Bind...";
                tcpListener = null;
                ipAddressServerBind = IPAddress.Any;            ///< Any available IP.
                try
                {
                    tcpListener = new TcpListener(ipAddressServerBind, serverPortSelect);
                    tcpListener.Start();                        ///< Starts server.
                    success = true;
                }
                catch (SocketException exception)
                {
                    serverConnectionDetails.Text += Environment.NewLine + "Failed. Socket error Code = " + (exception.ErrorCode).ToString();
                }
                catch (ArgumentOutOfRangeException)
                {
                    serverConnectionDetails.Text += Environment.NewLine + "Failed. Unable to bind on this port.";
                }
                catch
                {
                    serverConnectionDetails.Text += Environment.NewLine + "Failed. Unable to bind.";
                }
                serverConnectButton.Enabled = true;
                if (success == false)
                {
                    serverPortNumber.Enabled = true;
                    serverConnectButton.Text = "Listen";
                    tcpListener = null;
                    return;
                }
                hostName = Dns.GetHostName();
                /// Get IPv4 address server is bound to.
                for (addressCount = 0; addressCount <= System.Net.Dns.GetHostEntry(hostName).AddressList.Length - 1; addressCount++)
                {
                    if (System.Net.Dns.GetHostEntry(hostName).AddressList[addressCount].AddressFamily == AddressFamily.InterNetwork)
                    {
                        serverConnectionDetails.Text += Environment.NewLine + "Bound to " +
                                                Dns.GetHostEntry(hostName).AddressList[addressCount].ToString() + ":" +
                                                serverPortNumber.Text + Environment.NewLine + "Listening for a connection...";
                        break;
                    }
                }
                serverConnectButton.Text = "Close";
                serverSendButton.Enabled = true;
                timerServerScan.Enabled = true;
            }
            else if (serverConnectButton.Text == "Close")
            {
                timerServerScan.Enabled = false;
                if (serverConnected == true)
                {
                    networkStreamConnectedClient.Close();       ///< Closes client network stream.
                    tcpClientOnServer.Close();                  ///< Closes client socket.
                }
                tcpListener.Stop();                             ///< Stops server.
                networkStreamConnectedClient = null;
                tcpClientOnServer = null;
                tcpListener = null;
                ipAddressServerBind = null;
                serverConnected = false;
                serverConnectButton.Text = "Listen";
                serverConnectionDetails.Text = "Disconnected.";
                serverSentBytes.Text = "0";
                serverSendButton.Enabled = false;
                serverReceivedBytes.Text = "Empty";
                serverPortNumber.Enabled = true;
            }
        }

        private void serverSendButton_Click(object sender, EventArgs e)
        {
            Int32 totalNumberOfBytes = 0;
            System.Text.UTF8Encoding encoding = new System.Text.UTF8Encoding();
            serverSentBytes.Text = null;
            totalNumberOfBytes = serverDataToSend.Text.Length;
            encoding.GetBytes(serverDataToSend.Text, 0, totalNumberOfBytes, serverTransmitBuffer, 0);
            try
            {
                networkStreamConnectedClient.Write(serverTransmitBuffer, 0, totalNumberOfBytes);        ///< Sends data.
            }
            catch (SystemException)
            {
                serverSentBytes.Text = "0";
                return;
            }
            serverSentBytes.Text = Convert.ToString(totalNumberOfBytes);
        }

        private void timerServerScan_Tick(object sender, EventArgs e)
        {
            IPEndPoint clientConnected = null;
            Int32 totalBytesReceived = 0;
            /// Checks for connection request from client, if already not connected.
            if (tcpListener.Pending() && serverConnected == false)
            {
                tcpClientOnServer = new TcpClient();
                tcpClientOnServer = tcpListener.AcceptTcpClient();
                clientConnected = (IPEndPoint)tcpClientOnServer.Client.RemoteEndPoint;
                serverConnectionDetails.Text += Environment.NewLine + "Connected to " + IPAddress.Parse(clientConnected.Address.ToString());
                try
                {
                    networkStreamConnectedClient = tcpClientOnServer.GetStream();       ///< Gets network stream for Read/Write.
                }
                catch
                {
                    serverConnectionDetails.Text += Environment.NewLine + 
                                                    IPAddress.Parse(clientConnected.Address.ToString()) +
                                                    "Send/Receive failure.";
                }
                serverConnectionDetails.SelectionStart = serverConnectionDetails.Text.Length;
                serverConnectionDetails.ScrollToCaret();
                serverSendButton.Enabled = true;
                serverConnected = true;
            }
            /// Checks if connection is active and scans for data from client side.  
            if (serverConnected == true)
            {
                if (tcpClientOnServer.Client.Poll(100, SelectMode.SelectRead))
                {
                    if (networkStreamConnectedClient.CanRead)
                    {
                        if (networkStreamConnectedClient.DataAvailable)
                        {
                            try
                            {
                                /// Reads data from connected client, if available.
                                totalBytesReceived = networkStreamConnectedClient.Read(serverReceiveBuffer, 0, bufferSize);
                            }
                            catch (SystemException)
                            {
                                serverReceivedBytes.Text = "Fail";
                                return;
                            }
                            serverReceivedBytes.Text = totalBytesReceived.ToString();
                            serverDataReceived.Text = Encoding.UTF8.GetString(serverReceiveBuffer, 0, totalBytesReceived);
                            serverDataReceived.SelectionStart = serverDataReceived.Text.Length;
                            serverDataReceived.ScrollToCaret();
                        }
                        else
                        {
                            /// Sends signal to serverConnectButton to disconnect the socket as
                            /// it is no more active.
                            serverConnectButton_Click(sender, e);
                        }
                    }
                }
            }
        }

        private void serverSendClearButton_Click(object sender, EventArgs e)
        {
            serverDataToSend.Text = null;
            serverSentBytes.Text = "0";
        }

        private void serverReceiveClearButton_Click(object sender, EventArgs e)
        {
            serverDataReceived.Text = null;
            serverReceivedBytes.Text = "Empty";
        }

        /// Procedure to update the form controls from other thread.
        private void UpdateClientFormControls(string totalBytes, string dataReceived, bool clientConnectButton)
        {
            object sender = null;
            if (InvokeRequired)
            {
                this.Invoke(new Action<string, string, bool>(UpdateClientFormControls), new object[] { totalBytes, dataReceived, clientConnectButton });
                return;
            }
            clientReceivedBytes.Text = totalBytes;
            clientDataReceived.Text = dataReceived;
            clientDataReceived.SelectionStart = clientDataReceived.Text.Length;
            clientDataReceived.ScrollToCaret();
            if (clientConnectButton == true)
            {
                clientConnectButton_Click(sender, EventArgs.Empty);
            }
        }

        private void ClientReceiveProcedure(Socket clientSocket)
        {
            Int32 totalBytesReceived = 0;
            string totalBytesStringFormat = null;
            string dataReceivedStringFormat = null;
            while (true)
            {
                try
                {
                    /// Receives over asynchronous socket.
                    totalBytesReceived = clientSocket.Receive(clientReceiveBuffer, bufferSize, SocketFlags.None);
                }
                catch
                {
                    UpdateClientFormControls("Fail", null, false);
                    break;
                }
                if (clientDisconnectCommand == true)
                {
                    break;
                }
                if (totalBytesReceived == 0)
                {
                    UpdateClientFormControls("Empty", null, true);
                    break;
                }
                totalBytesStringFormat = totalBytesReceived.ToString();
                dataReceivedStringFormat = Encoding.UTF8.GetString(clientReceiveBuffer, 0, totalBytesReceived);
                UpdateClientFormControls(totalBytesStringFormat, dataReceivedStringFormat, false);
            }
        }

        private void clientConnectButton_Click(object sender, EventArgs e)
        {
            bool success = false;
            LingerOption lingeroption = null;
            if (clientConnectButton.Text == "Connect")
            {
                if (ValidateIPAddress(clientConnectIPAddress.Text) == false)
                {
                    clientConnectionDetails.Text = "Enter valid IP address";
                    return;
                }
                try
                {
                    ipAddressClientConnect = IPAddress.Parse(clientConnectIPAddress.Text);      ///< Checks for valid IP.
                }
                catch
                {
                    clientConnectionDetails.Text = "Enter valid IP address";
                    return;
                }
                try
                {
                    /// Checks if port number is valid.
                    clientPortSelect = (Int32)Convert.ToUInt32(clientPortNumber.Text);
                    if ((clientPortSelect > 65535) || (clientPortSelect < 1))
                    {
                        throw new ArgumentOutOfRangeException();
                    }
                }
                catch
                {
                    clientConnectionDetails.Text = "Port number must be positive value less than 65535.";
                    return;
                }
                clientConnectionDetails.Text = "Trying to connect...";
                clientConnectButton.Text = "Trying...";
                clientConnectButton.Enabled = false;
                clientPortNumber.Enabled = false;
                clientConnectIPAddress.Enabled = false;
                try
                {
                    clientSocketEndPoint = new IPEndPoint(ipAddressClientConnect, clientPortSelect);
                    tcpClientSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    lingeroption = new LingerOption(false, 0);
                    tcpClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.Linger, lingeroption);
                    tcpClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.DontLinger, false);
                    clientSocketConnect = tcpClientSocket.BeginConnect(clientSocketEndPoint, null, null);
                    /// Sets time-out for BeginConnect procedure.
                    success = clientSocketConnect.AsyncWaitHandle.WaitOne(timeoutMsec, false);
                    if (success == false)
                    {
                        clientConnectionDetails.Text += Environment.NewLine + "Timeout.";
                        tcpClientSocket.Close();
                    }
                }
                catch (SocketException exception)
                {
                    clientConnectionDetails.Text += Environment.NewLine + "Failed. Socket error Code = " + (exception.ErrorCode).ToString();
                }
                catch
                {
                    clientConnectionDetails.Text += Environment.NewLine + "Failed. Unable to connect.";
                }
                clientConnectButton.Enabled = true;
                if (success == false)
                {
                    tcpClientSocket = null;
                    clientConnectButton.Text = "Connect";
                    clientPortNumber.Enabled = true;
                    clientConnectIPAddress.Enabled = true;
                    return;
                }
                clientConnectButton.Text = "Disconnect";
                clientConnectionDetails.Text += Environment.NewLine + "Connected.";
                clientSendButton.Enabled = true;
                /// New thread to watch reception of data and connectivity.
                threadClientSide = new Thread(() => ClientReceiveProcedure(tcpClientSocket));
                threadClientSide.Start();
                clientDisconnectCommand = false;
            }
            else if (clientConnectButton.Text == "Disconnect")
            {
                clientDisconnectCommand = true;
                try
                {
                    tcpClientSocket.Shutdown(SocketShutdown.Both);      ///< Stops Send/Receive immediately.
                    Thread.Sleep(50);
                    tcpClientSocket.Close();                            ///< Closes the socket.
                }
                catch
                {
                    clientConnectionDetails.Text += Environment.NewLine + "No connection exists.";
                }
                threadClientSide = null;
                tcpClientSocket = null;
                clientSocketEndPoint = null;
                clientConnectButton.Text = "Connect";
                clientConnectionDetails.Text += Environment.NewLine + "Disconnected.";
                clientSendButton.Enabled = false;
                clientSentBytes.Text = "0";
                clientReceivedBytes.Text = "Empty";
                clientPortNumber.Enabled = true;
                clientConnectIPAddress.Enabled = true;
            }
        }

        private void clientSendButton_Click(object sender, EventArgs e)
        {
            Int32 totalNumberOfBytes = 0;
            System.Text.UTF8Encoding encoding = new System.Text.UTF8Encoding();
            clientSentBytes.Text = null;
            totalNumberOfBytes = clientDataToSend.Text.Length;
            encoding.GetBytes(clientDataToSend.Text, 0, totalNumberOfBytes, clientTransmitBuffer, 0);
            try
            {
                tcpClientSocket.Send(clientTransmitBuffer, totalNumberOfBytes, SocketFlags.None);       ///< Sends data.
            }
            catch
            {
                clientSentBytes.Text = "Fail";
                return;
            }
            clientSentBytes.Text = Convert.ToString(totalNumberOfBytes);
        }

        private void ClientSendClearButton_Click(object sender, EventArgs e)
        {
            clientDataToSend.Text = null;
            clientSentBytes.Text = "0";
        }

        private void clientReceiveClearButton_Click(object sender, EventArgs e)
        {
            clientDataReceived.Text = null;
            clientReceivedBytes.Text = "Empty";
        }

        private void Form1_Closing(object sender, CancelEventArgs e)
        {
            /// Disconnects and stops the server, if active.
            if (serverConnectButton.Text != "Listen")
            {
                serverConnectButton_Click(sender, e);
            }
            /// Disconnects the Client.
            if (clientConnectButton.Text != "Connect")
            {
                clientConnectButton_Click(sender, e);
            }
        }
    }
}